// $Id: dekker.cpp,v 1.5 2006/09/17 22:27:22 sharma Exp $
//An implementation of the Dekker-Brent for approximating roots in a given
//interval. The worst case behaviour of this procedure is similar to bisection,
//but the best behaviour is super-linear convergence as in Secant method.

#define CORE_LEVEL 4
#include "CORE/CORE.h"
#include "readpoly.h"

typedef BigInt NT;
typedef Polynomial<NT> Poly;

using namespace CORE;
using namespace std;

// This is the algorithm suggested by Dekker, without the modifications
//by Brent.
//
//The input is an interval J such that P has opposite signs at the endpoints
//of J. The output is an
//interval of width less than 2^{-aprec} at whose endpoints P has opposite
//signs. We assume that there is a unique root of P in J and that the
//endpoints of J are exact BigFLoats. 

BFInterval dekker(Polynomial<NT>& P, BFInterval J, int aprec)
{
  if((J.second - J.first).uMSB() < -aprec)
    return (J);
  
  BigFloat i;
  BigFloat a = J.first, b=J.second, c=a, m;
  // b is our best approxmitaion to the root.
  // a is the value of b at the previous iteration.
  // c is the contrapoint, i.e. fb * fc < 0 always.
  // m is (b+c)/2.
  // i is the interpolant between (a, fa) and (b, fb).
  BigFloat fa, fb, fc, fi;
  BigFloat tol = BigFloat(1, -aprec);// tol = 2^{-aprec}
  BigFloat temp;

  
  //Initializations:
  fa=P.eval(a); fb=P.eval(b); fc = fa;
  //Check if a is a root
  if(sign(fa) == 0)
      return BFInterval(a,a);
  //Check if b is a root
  if(sign(fb) == 0)
      return BFInterval(b,b);

  if(core_abs(fc) > core_abs(fb)){//if c is closer to root then b is, swap them
    temp= b; b=c; c = temp;
    temp = fb; fb= fc; fc = temp;
  }
  while( core_abs(b-c).uMSB() >= -aprec){
    m = (b+c).div2();

    if(fb != fa){
      i = (a*fb - b*fa)/(fb-fa);// the linear interpolation/extrapolation
                            // What is the precision with which to do division?
    }else
      i = m;//have to do a bisection since extrapolation does not work

    i.makeExact();
    if(core_min(b,c) > i || i > core_max(b,c))// i.e. we had an extrapolation
      i = m;    //and i is not between b and c, then we reset i to (b+c)/2.

    fi = P.eval(i);

    if(core_abs(b-i).uMSB() < -aprec){// b and i are too close
      if(sign(fb)*sign(fi) < 0)
	return BFInterval(core_min(b,i), core_max(b,i));
      i = b + sign(c-b)*tol;// thus |b-i| = tol
      fi = P.eval(i);
      if(sign(fb)*sign(fi) < 0)
	return BFInterval(core_min(b,i), core_max(b,i));
    }else{//|b-i| >= tol
      if( core_min(m, b) > i || i > core_max(m, b)){
	i = m;// i is not in between b and m, so reset it to m.
	fi = P.eval(i);
      }
    }
    //Check if i is a root
    if(sign(fi) == 0)
      return BFInterval(i,i);

    //Make the assignments for the next iteration
    a = b; fa = fb; b=i; fb = fi;
    
    if(fb * fc < 0){
      //Then fa*fb > 0, so in the next iteration we will have an extrapolation
      //unless the following happens
      if(core_abs(fc) < core_abs(fb)){//if c is closer to root then b is
	temp= b; b=c; c = temp;       //swap them.
	temp = fb; fb= fc; fc = temp;
      }
    }else{//fb * fc > 0. Then fa*fb < 0, and we reset c to a, to maintain
      // the invariant fb*fc < 0.      
      c = a; fc = fa;
    }
    cout << "New interval " << b << ":" << c << endl;
    cout << "Width is "<< core_abs(b-c).uMSB() << endl;
    
  }

  if((core_max(b,c) -core_min(b,c)).uMSB() >= -aprec)
    cout << "ERROR! Interval not refined" << endl;

  return BFInterval(core_min(b,c), core_max(b,c));
  
}//dekker


// The following code is the one given by Dekker in his paper, and 
// is perhaps efficient then the one above.
BFInterval dekker1(Polynomial<NT>& P, BFInterval J, int aprec)
{
  if((J.second - J.first).uMSB() < -aprec)
    return (J);
  
  BigFloat p, q;
  BigFloat a = J.first, b=J.second, c=a, m;
  // b is our best approxmitaion to the root.
  // a is the value of b at the previous iteration.
  // c is the contrapoint, i.e. fb * fc < 0 always.
  // m is (b+c)/2.
  // (p/q) +b is the interpolant between (a, fa) and (b, fb).
  BigFloat fa, fb, fc, fi;
  BigFloat tol = BigFloat(1, -aprec);// tol = 2^{-aprec}
  BigFloat temp;

  
  //Initializations:
  fa=P.eval(a); fb=P.eval(b); fc = fa;
  //Check if a is a root
  if(sign(fa) == 0)
      return BFInterval(a,a);
  //Check if b is a root
  if(sign(fb) == 0)
      return BFInterval(b,b);

  if(core_abs(fc) > core_abs(fb)){//if c is closer to root then b is, swap them
    temp= b; b=c; c = temp;
    temp = fb; fb= fc; fc = temp;
  }
  while( core_abs(b-c).uMSB() >= -aprec){
    m = (b+c).div2();

    /*    if(fb != fa){
      i = (a*fb - b*fa)/(fb-fa);// the linear interpolation/extrapolation
                            // What is the precision with which to do division?
    }else
      i = m;//have to do a bisection since extrapolation does not work
    */
    p = (b-a)*fb; q = fa - fb;//Some problem with q? throws an out_of_range flag
    a = b; fa = fb;

    // Make an assignment to b
    if(core_abs(p) <= core_abs(q)*tol){//This is equivalent to |b-i| <= tol
      b += sign(c-b)*tol;// where i is defined as in the above procedure.
    }else{
      if(core_abs(p) < (m-b)*core_abs(q)){//This means i is between b and m.
	cout << "p= " << p <<" q= " << q << endl;
	b += (p/q).makeExact();
      }else{
	b = m;
      }
    }
    fb = P.eval(b);

    //Check if b is a root
    if(sign(fb) == 0)
      return BFInterval(b,b);

    if(fb * fc < 0){
      //Then fa*fb > 0, so in the next iteration we will have an extrapolation
      //unless the following happens
      if(core_abs(fc) < core_abs(fb)){//if c is closer to root then b is
	temp= b; b=c; c = temp;       //swap them.
	temp = fb; fb= fc; fc = temp;
      }
    }else{//fb * fc > 0. Then fa*fb < 0, and we reset c to a, to maintain
      // the invariant fb*fc < 0.      
      c = a; fc = fa;
    }
    cout << "New interval " << b << ":" << c << endl;
    cout << "Width is "<< core_abs(b-c).uMSB() << endl;
    
  }

  if((core_max(b,c) -core_min(b,c)).uMSB() >= -aprec)
    cout << "ERROR! Interval not refined" << endl;

  return BFInterval(core_min(b,c), core_max(b,c));
  
}//dekker1

//This code incorporates Brent's modifications to Dekker's original
// algorithm. The ALGOL equivalent is described in Brent's book
// Algorithms for Minimization Without Derivatives, Page 58.
BFInterval dekkerBrent1(Polynomial<NT>& P, BFInterval J, int aprec)
{
  if((J.second - J.first).uMSB() < -aprec)
    return (J);
  
  BigFloat p, q, r, s;
  BigFloat a = J.first, b=J.second, c=a, m, d, e;
  // b is our best approxmitaion to the root.
  // a is the value of b at the previous iteration.
  // c is the contrapoint, i.e. fb * fc < 0 and |fb| <= |fc| always.
  // m is (c-b)/2.
  // (p/q)+b is the linear interpolant between (a, fa) and (b, fb).
  // p, q, r and s are used in the inverse quadratic interpolant
  // e is the distance between b and c at the iteration before the previous one.
  // d is the distance between b and the new iterate
  BigFloat fa, fb, fc, fi;
  BigFloat tol = BigFloat(1, -aprec);// tol = 2^{-aprec}
  BigFloat temp;

  //Initializations:
  fa=P.eval(a); fb=P.eval(b); fc = fa;
  //Check if a is a root
  if(sign(fa) == 0)
      return BFInterval(a,a);
  //Check if b is a root
  if(sign(fb) == 0)
      return BFInterval(b,b);

  if(core_abs(fc) > core_abs(fb)){//if c is closer to root then b is, swap them
    temp= b; b=c; c = temp;
    temp = fb; fb= fc; fc = temp;
  }

  while( core_abs(b-c).uMSB() >= -aprec){//while the desired precision is not attained
    m = (c-b).div2();
    if(core_abs(e).uMSB() >= -aprec && core_abs(fa) > core_abs(fb)){
      //That is, the previous step was big and in the right direction
      //so interpolation can be tried in this step.
      s = (fb/fa).makeExact();
      if(a == c){//Linear interpolation
	p = 2*m*s; q = 1-s;
      }else{//Inverse Quadratic Interpolation (IQI)
	q = (fa/fc).makeExact(); r = (fb/fc).makeExact();
	p = s*(2*m*q*(q-r)- (b-a)*(r-1));
	q = (q-1)*(r-1)*(s-1);
      }
      if(p > 0) 
	q *=-1;
      else
	p *= -1;
      s = e; e = d;
      if (2*p < 3 *m*q - core_abs(tol * q) && 2*p < core_abs(s*q))
	d = (p/q).makeExact();// The first condition means that b+p/q
      // is between b and c.
      // The second means that |b-i|, i = b+(p/q), is smaller than half
      // the gap at the iteration before the previous one. This was the
      // main modification suggested by Brent.
      else
	d = e= m;
    }else
            d = e= m;

    a = b; fa = fb;
    if(core_abs(d).uMSB() > -aprec)//If the next possible iterate is
      b += d;                      // sufficiently far.
    else
      b += sign(m)*tol;

    fb = P.eval(b);
    
    //Check if b is a root
    if(sign(fb) == 0)
      return BFInterval(b,b);

    if(fb * fc < 0){
      //Then fa*fb > 0, so in the next iteration we will have an extrapolation
      //unless the following happens
      if(core_abs(fc) < core_abs(fb)){//if c is closer to root then b is
	temp= b; b=c; c = temp;       //swap them.
	temp = fb; fb= fc; fc = temp;
      }
    }else{//fb * fc > 0. Then fa*fb < 0, and we reset c to a, to maintain
      // the invariant fb*fc < 0.      
      c = a; fc = fa;
    }
    cout << "New interval " << b << ":" << c << endl;
    cout << "Width is "<< core_abs(b-c).uMSB() << endl;
    
  }//end of main while

  if((core_max(b,c) -core_min(b,c)).uMSB() >= -aprec)
    cout << "ERROR! Interval not refined" << endl;

  return BFInterval(core_min(b,c), core_max(b,c));
  
}//dekkerBrent1


//The following is a code exported from Netlib, but modified to suit
//Core Library; basically, changing doubles to BigFloats and ensuring
//that after division there are no errors.
//The original code is available from
//                   http://www.netlib.org/c/brent.shar.
//Here is a description of the algorithm borrowed from the same place.
/* Algorithm
 *	G.Forsythe, M.Malcolm, C.Moler, Computer methods for mathematical
 *	computations. M., Mir, 1980, p.180 of the Russian edition
 *
 *	The function makes use of the bissection procedure combined with
 *	the linear or quadric inverse interpolation.
 *	At every step program operates on three abscissae - a, b, and c.
 *	b - the last and the best approximation to the root
 *	a - the last but one approximation
 *	c - the last but one or even earlier approximation than a that
 *		1) |f(b)| <= |f(c)|
 *		2) f(b) and f(c) have opposite signs, i.e. b and c confine
 *		   the root
 *	At every step Zeroin selects one of the two new approximations, the
 *	former being obtained by the bissection procedure and the latter
 *	resulting in the interpolation (if a,b, and c are all different
 *	the quadric interpolation is utilized, otherwise the linear one).
 *	If the latter (i.e. obtained by the interpolation) point is 
 *	reasonable (i.e. lies within the current interval [b,c] not being
 *	too close to the boundaries) it is accepted. The bissection result
 *	is used in the other case. Therefore, the range of uncertainty is
 *	ensured to be reduced at least by the factor 1.6
 */
/* ASSERT: J has exact endpoints and P has different signs at the endpoints of J*/

BFInterval dekkerBrent(Polynomial<NT>& P, BFInterval J, int aprec)
{
  if((J.second - J.first).uMSB() < -aprec)
    return (J);


  BigFloat a,b,c;			/* descr. see above	*/
  BigFloat fa;				/* f(a)				*/
  BigFloat fb;				/* f(b)				*/
  BigFloat fc;				/* f(c)				*/

  a = J.first;  b = J.second;  fa = P.eval(a);  fb = P.eval(b);
  //Check if a is a root
  if(sign(fa) == 0)
      return BFInterval(a,a);
  //Check if b is a root
  if(sign(fb) == 0)
      return BFInterval(b,b);

  assert(fa*fb < 0);

  c = a;   fc = fa;

  BigFloat prev_step;     		/* Distance from the last but one*/
					/* to the last approximation	*/
  BigFloat tol_act=BigFloat(1,-aprec);	/* Actual tolerance		*/
  BigFloat p;      			/* Interpolation step is calcu- */
  BigFloat q;      			/* lated in the form p/q; divi- */
  					/* sion operations is delayed   */
 					/* until the last moment	*/
  BigFloat new_step;      		/* Step at this iteration       */

  BigFloat t1,cb,t2;

  for(;;)		/* Main iteration loop	*/
  {
     prev_step = b-a;    		/* Distance from the last but one*/
					/* to the last approximation	*/
   
    if( core_abs(fc) < core_abs(fb) )
    {                         		/* Swap data for b to be the 	*/
	a = b;  b = c;  c = a;          /* best approximation		*/
	fa=fb;  fb=fc;  fc=fa;
    }
    //    tol_act = 2*EPSILON*core_abs(b) + tol/2;
    new_step = (c-b).div2();

    if( core_abs(new_step).uMSB() < -aprec)      /* Reached the desired precision */
      return BFInterval(core_min(b,c), core_max(b,c));
    if( fb == 0 )                       /* Found the exact root */
      return BFInterval(b, b);				

    			/* Decide if the interpolation can be tried	*/
    if( core_abs(prev_step).uMSB() >= -aprec	/* If prev_step was large enough*/
	&& core_abs(fa) > core_abs(fb) )	/* and was in true direction,	*/
    {					/* Interpolatiom may be tried	*/
      
	cb = c-b;
	if( a==c )			/* If we have only two distinct	*/
	{				/* points linear interpolation 	*/
	  t1 = (fb/fa).makeExact();	/* can only be applied		*/
	  p = cb*t1;
	  q = 1.0 - t1;
 	}
	else				/* Quadric inverse interpolation*/
	{
	  q = (fa/fc).makeExact();  
	  t1 = (fb/fc).makeExact();  
	  t2 = (fb/fa).makeExact();
	  p = t2 * ( cb*q*(q-t1) - (b-a)*(t1-1.0) );
	  q = (q-1.0) * (t1-1.0) * (t2-1.0);
	}
	if( p>0 )		/* p was calculated with the op-*/
	  q = -q;			/* posite sign; make p positive	*/
	else				/* and assign possible minus to	*/
	  p = -p;			/* q				*/

	if( p < (0.75*cb*q-core_abs(tol_act*q).div2())/* If b+p/q falls in [b,c]*/
	    && 2*p < core_abs(prev_step*q) )	/* and isn't too large	*/
	  new_step = (p/q).makeExact();			/* it is accepted	*/
					/* If p/q is too large then the	*/
					/* bissection procedure can 	*/
					/* reduce [b,c] range to more	*/
					/* extent			*/
    }

    if( core_abs(new_step).uMSB() < -aprec )/* Adjust the step to be not less*/
      if( new_step > 0 )        	/* than tolerance		*/
	new_step = tol_act;
      else
	new_step = -tol_act;

    a = b;  fa = fb;			/* Save the previous approx.	*/
    b += new_step;  fb = P.eval(b);	/* Do step to a new approxim.	*/
    if( (fb > 0 && fc > 0) || (fb < 0 && fc < 0) )
    {                 			/* Adjust c for it to have a sign*/
      c = a;  fc = fa;                  /* opposite to that of b	*/
    }
    //    cout << "New interval " << b << ":" << c << endl;
    //    cout << "Width is "<< core_abs(b-c).uMSB() << endl;

  }

}//dekkerBrent



int main( int argc, char* argv[] ) {
  
  //  for(int i=0; i <= deg; i++)
  //    cout << i <<"th coefficient is "<< temp[i] << endl;

  Poly P;
  
  /* ************************************** */
  /* Read Arguments */
  /* ************************************** */
  if (argc > 1)
	read_poly(argv[1], P);
  else
    read_poly("../data/polynomials/wilk20.pol", P);
  //read_poly("../data/polynomials/hermite20.pol", P);
  //read_poly("../data/polynomials/geom3_20.pol", P);

  BFInterval I=BFInterval(1.5, 2.9);
  if(argc > 4){
    I.first = BigFloat(argv[2]).makeFloorExact();
    I.second = BigFloat(argv[3]).makeCeilExact();
  }
  int prec = 54;
  if(argc == 5)
    prec = atoi(argv[4]);

  cout <<"Desired precision of " << prec << endl;
  Timer t1, t2;
  t1.start();
  BFInterval J = dekkerBrent(P, I, prec);
  t1.stop();
  cout << "Time taken by Dekker-Brent " << t1.getSeconds() << endl;
  cout <<"Refined interval: "<< J.first << ":" << J.second << endl;
  if(J.second == J.first)
    cout << "Got the exact root " << J.first << endl;
  else{
    if((J.second -J.first).uMSB() > -prec)
      cout << "ERROR! Interval not refined. Width is:" << (J.second -J.first).uMSB() << endl;
  }

  Sturm<NT> SS(P);
  t2.start();
  J = SS.newtonRefine(I, prec);
  t2.stop();
  cout << "Time taken by Newton-refine " << t2.getSeconds() << endl;  
  cout <<"Refined interval from Newton: "<< J.first << ":" << J.second << endl;
    
  return 0;
}

/* A table comparing the timings of Dekker-Brent against our newtonRefine.

Polynomial    Input Interval         Prec    Dekker-Brent  Newton-refine
-------------------------------------------------------------------------
lsr_24       [8.58e+9,1.72e+11]      10      0.01          0.01
                                     100     0.01          0.02
                                     1000    0.04          0.04
                                     5000    0.59          1.00
                                     10000   3.26          6.69

wilk160      [99.5,100.9]            10      0.02          0.14
                                     100     0.04          0.22
                                     1000    0.61          2.74
                                     5000    27.54         74.19

laguerre40   [104,112.5]             10      0.01          0.02
                                     100     0.01          0.03
                                     1000    0.07          0.11
                                     5000    1.64          3.71

chebyshev80  [-0.15625,-0.12500]     10      0.01          0.02
                                     100     0.01          0.03
                                     1000    0.21          0.21
                                     5000    6.19          15.89

hermite80    [-11. ; -10.5000]       10      0.01          0.04
                                     100     0.02          0.06
                                     1000    0.23          0.25
                                     5000    6.35          15.67

*/
